library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use ieee.std_logic_unsigned.all;


entity receiver is
 generic(
       firLength: natural := 135;
       maxComLength: natural :=1023;
       maxALBits: natural:=14;
       maxAL: natural:=16384
     );
    port(CLK160: in std_logic;
         CLK80: in std_logic;
         CLK20: in std_logic;
         RST: in std_logic;
         RUNQ: in std_logic;
         ADCLK: in std_logic;
         AQC: in std_logic;
         AQS: in std_logic;
         SIG: in signed(13 downto 0);
         ACQ_PHASE: in std_logic_vector(1 downto 0);
         ACQ_START: in std_logic;
         RXF_N : in std_logic;
         TXE_N : in std_logic;
         RD_N  : out std_logic;
         WR    : out std_logic;
         USBDATA: inout std_logic_vector(7 downto 0);
         ACQ_BUSY: out std_logic;
         TRANS_BUSY: out std_logic;
         RG: in std_logic
         );
end receiver;


architecture RTL of receiver is

  component quadDemodulator is
    port(CLK: in std_logic;
         RST: in std_logic; 
         SIG: in signed(13 downto 0);
         TRIG: out std_logic;
         SIG_C, SIG_S: out signed(13 downto 0)
         );
  end component;



  component signalAccumulator is
   generic(
       maxALBits: natural
     );
    port(CLK: in std_logic;
         RST: in std_logic;
         RUNQ: in std_logic;
         ACQ_PHASE: in std_logic_vector(1 downto 0);
         ACQ_START: in std_logic;
         SIG_C: in signed(22 downto 0);
         SIG_S: in signed(22 downto 0);
--         ASIG_C: in std_logic_vector(27 downto 0); --
--         ASIG_S: in std_logic_vector(27 downto 0); --
         ASIG_C: in std_logic_vector(31 downto 0); --
         ASIG_S: in std_logic_vector(31 downto 0); --
         DWN: in std_logic_vector(15 downto 0); 
         
         AL: in std_logic_vector(maxALBits-1 downto 0);
         NA: in std_logic_vector(31 downto 0);
         STEP: in std_logic_vector(15 downto 0);
         TRANS_ACTIVE: in std_logic;
         ACQ_ACTIVE: out std_logic;
  --       CA: out std_logic_vector(31 downto 0);
--         QSIG_C: out std_logic_vector(27 downto 0); --
--         QSIG_S: out std_logic_vector(27 downto 0); --
         QSIG_C: out std_logic_vector(31 downto 0); --
         QSIG_S: out std_logic_vector(31 downto 0); --

         ADDRESS: out std_logic_vector(maxALBits-1 downto 0);
         WE: out std_logic;
         ACCUM_OK: out std_logic
         );
  end component;



  component usbArbiter is
    port ( CLK : in  std_logic;
            RST : in  std_logic;
            RXF_N : in std_logic; 
            TXE_N : in std_logic;
            RD_N  : out std_logic;
            WR    : out std_logic;
            BUS_BUSY : out std_logic;
            TG_READY : in  std_logic;
            USB_DATA : inout  std_logic_vector( 7 downto 0 );
            TG_DO    : out std_logic_vector( 7 downto 0 );
            TG_DI    : in  std_logic_vector( 7 downto 0 );
            TG_RD    : out std_logic;
            TG_WR    : in  std_logic );
  end component;


  component hexToA
    port( CLK, RST, G: in std_logic;
          D: in std_logic_vector(3 downto 0);
          Q: out std_logic_vector(7 downto 0)
         );
  end component;

  component aToHex
    port( CLK, RST, G: in std_logic;
          D: in std_logic_vector(7 downto 0);
          Q: out std_logic_vector(3 downto 0)
         );
  end component;

--------- variables for quadrature demodulator ---------

  signal QDStateReg: std_logic_vector(1 downto 0):="00";
  signal sigReg: signed(13 downto 0);
  signal sigCReg, sigSReg: signed(13 downto 0);
  signal sigCLatchReg,sigSLatchReg: std_logic;

--------- variables for averager ---------

  signal avCStateReg, avSStateReg: std_logic_vector(4 downto 0):="00000";
  signal DFTrigCReg, DFTrigSReg: std_logic;
 
  type avSig is array (0 to 9) of signed(17 downto 0);
  signal avSigSReg,avSigCReg: avSig;
  
--------- variables for digital filter ---------
  type coeff is array(0 to firLength-1) of signed(9 downto 0);
  signal DFcoeff: coeff;

  type DFQ is array(0 to firLength-1) of signed(23 downto 0);
  signal CQRealReg, CQImagReg: DFQ;

  type DFS is array(0 to firLength-1) of signed(22 downto 0);
  signal CSRealReg, CSImagReg: DFS;

  signal dfLengthReg: integer range 0 to firLength-1; 
------------------------------------------------


--  type FID is array(0 to maxAL-1) of std_logic_vector(27 downto 0);
  type FID is array(0 to maxAL-1) of std_logic_vector(31 downto 0);

  signal FID_real, FID_imag: fid;

  signal qdTrigReg, C50nsLatchReg, S50nsLatchReg: std_logic;
  
  signal sigCReg2a,sigCReg2b,sigSReg2a,sigSReg2b: signed(22 downto 0);
  
  signal sigCReg3, sigSReg3: signed(22 downto 0);
 
  signal datBuf32Reg: std_logic_vector(31 downto 0);
  signal add1Reg, add2Reg, addQReg: std_logic_vector(maxALBits-1 downto 0);
  signal alReg: std_logic_vector(31 downto 0);
  signal dwReg: std_logic_vector(31 downto 0);
  signal naReg: std_logic_vector(31 downto 0);
  signal stepReg: std_logic_vector(31 downto 0);
  signal afReg: std_logic_vector(31 downto 0);
  signal acqActiveReg: std_logic;
--  signal AsigCReg, AsigSReg: std_logic_vector(27 downto 0);
--  signal QsigCReg, QsigSReg: std_logic_vector(27 downto 0);
  signal AsigCReg, AsigSReg: std_logic_vector(31 downto 0);
  signal QsigCReg, QsigSReg: std_logic_vector(31 downto 0);
--  signal caReg: std_logic_vector(31 downto 0);
  signal accumOKReg, transActiveReg, weReg: std_logic;
  signal readyCReg, readySReg: std_logic;
  signal diReg, doReg: std_logic_vector(7 downto 0);
  signal tgrdReg, tgwrReg, busBusyReg, readyReg: std_logic;

  signal QDRSTReg, QDRSTReg_SYNC, aReg, bReg: std_logic;
  signal samplingTriggerReg, stReg, nstReg: std_logic;
  
  signal coReg: std_logic_vector(9 downto 0);

  

-- variables for RCVR_interface --
  signal INTF_stateReg: std_logic_vector(7 downto 0):=x"00";
  type TSentence is array(0 to maxComLength) of std_logic_vector(7 downto 0);
  signal A: TSentence;
  signal messageReg: std_logic_vector(7 downto 0);
  signal bufLength: integer range 0 to maxComLength; -- keeps the length of A
  signal kReg, mReg: integer range 0 to maxComLength;  -- counter
  signal HADReg: std_logic_vector(3 downto 0);
  signal HAQReg: std_logic_vector(7 downto 0);
  signal AHDReg, AHD2Reg, AHD3Reg: std_logic_vector(7 downto 0);
  signal AHQReg, AHQ2Reg, AHQ3Reg: std_logic_vector(3 downto 0);
  signal RDPReg: std_logic_vector(15 downto 0);
  signal STPReg: std_logic_vector(23 downto 0);

-- end of the variables for RCVR_interface --

begin

  -----------------------   		
  --   Receiver gate   --
  -----------------------
  process(CLK80) begin
    if(CLK80'event and CLK80='1') then
      if (RG='0') then
        sigReg<=(others=>'0');
      else 
        sigReg<=SIG;
      end if;
    end if;
  end process;

---- detection of rising edge in the QDRST signal ----
  process(CLK80) begin
    if(CLK80'event and CLK80='1') then
      aReg <= QDRSTReg;
      bReg <= not aReg;
    end if;
  end process;

  QDRSTReg_SYNC <= aReg and bReg; 
-------------------------------------------------------


---- Main of quadrature demodulation ---------
  process(CLK80) begin
    if(CLK80'event and CLK80='1') then
      if QDRSTReg_SYNC='1' then QDstateReg <= "00";       
      else
        case QDStateReg is
          when "00" => 
			   sigCReg <= sigReg;
				sigCLatchReg <= '1';
				sigSLatchReg <= '0';
				C50nsLatchReg <= '1';
            QDStateReg <= "01";
          when "01" => 
			   sigSReg <= sigReg;
				sigCLatchReg <= '0';
				sigSLatchReg <= '1';
				S50nsLatchReg <= '1';
				QDStateReg <= "10";
          when "10" => 
			   sigCReg <= -sigReg;
				sigCLatchReg <= '1';
				sigSLatchReg <= '0';
				C50nsLatchReg <= '0';
				QDStateReg <= "11";
          when "11" => 
			   sigSReg <= -sigReg; 
				sigCLatchReg <= '0';
				sigSLatchReg <= '1';
				S50nsLatchReg <= '0';
				QDStateReg <= "00";
        end case;
      end if; -- RST;
    end if; -- CLK
  end process;

  
---- detection of the leading edge of sampling trigger ----
  process(CLK80) begin
    if(CLK80'event and CLK80='1') then
      stReg <= ACQ_START;
      nstReg <= not stReg;
    end if;
  end process;

  samplingTriggerReg <= stReg and nstReg;
-----------------------------------------------------------

  
  --
  --  [Signal averager]
  --
  --  This module receives the demodulated in-phase and quadrature signals,
  --  which is updated at a rate of 40 MHz. Here the signals are averaged 
  --  over 10 times, and sent to the digital filter. That is, the input signal 
  --  of the digital filter will be updated at a rate of 40/10=4 MHz.
  --
  process(CLK80) begin
    if(CLK80'event and CLK80='1') then
      if(samplingTriggerReg='1') then  -- synchronization with the sampling trigger
	     avCStateReg <= "00000"; 
      else
      case avCStateReg is  	  
	     when "00000"=> if(sigCLatchReg='1') then avCStateReg<="00001"; end if;
	     when "00001"=> if(sigCLatchReg='1') then avCStateReg<="00011"; end if;
	     when "00011"=> if(sigCLatchReg='1') then avCStateReg<="00111"; end if;
	     when "00111"=> if(sigCLatchReg='1') then avCStateReg<="01111"; end if;
	     when "01111"=> if(sigCLatchReg='1') then avCStateReg<="11111"; end if;
	     when "11111"=> if(sigCLatchReg='1') then avCStateReg<="11110"; end if;
	     when "11110"=> if(sigCLatchReg='1') then avCStateReg<="11100"; end if;
	     when "11100"=> if(sigCLatchReg='1') then avCStateReg<="11000"; end if;
	     when "11000"=> if(sigCLatchReg='1') then avCStateReg<="10000"; end if;
	     when "10000"=> if(sigCLatchReg='1') then avCStateReg<="00000"; end if;
	     when others => avCStateReg<="00000"; 
      end case;
		end if;
    end if;
  end process;
	       
  dfTrigCReg <= '1' when avCStateReg="10000" else '0';
    

  process(CLK80) begin
    if(CLK80'event and CLK80='1') then
      if(samplingTriggerReg='1') then  -- synchronization with the sampling trigger
	     avSStateReg <= "00000"; 
      else
      case avSStateReg is  	  
	     when "00000"=> if(sigSLatchReg='1') then avSStateReg<="00001"; end if;
	     when "00001"=> if(sigSLatchReg='1') then avSStateReg<="00011"; end if;
	     when "00011"=> if(sigSLatchReg='1') then avSStateReg<="00111"; end if;
	     when "00111"=> if(sigSLatchReg='1') then avSStateReg<="01111"; end if;
	     when "01111"=> if(sigSLatchReg='1') then avSStateReg<="11111"; end if;
	     when "11111"=> if(sigSLatchReg='1') then avSStateReg<="11110"; end if;
	     when "11110"=> if(sigSLatchReg='1') then avSStateReg<="11100"; end if;
	     when "11100"=> if(sigSLatchReg='1') then avSStateReg<="11000"; end if;
	     when "11000"=> if(sigSLatchReg='1') then avSStateReg<="10000"; end if;
	     when "10000"=> if(sigSLatchReg='1') then avSStateReg<="00000"; end if;
	     when others => avSStateReg<="00000"; 
      end case;
		end if;
    end if;
  end process;
	       
  dfTrigSReg <= '1' when avSStateReg="10000" else '0';
  
  process(CLK80)
  begin
    if (CLK80'event and CLK80='1') then
      if (sigCLatchReg='1') then
        avSigCReg(0) 
	       <= sigCReg(13) & sigCReg(13) & sigCReg(13) & sigCReg(13) & sigCReg;
        for I in 1 to 9 loop
	       avSigCReg(I) <= avSigCReg(I-1) 
		      + (sigCReg(13) & sigCReg(13) & sigCReg(13) & sigCReg(13) & sigCReg);
	     end loop;
      end if;
    end if;
  end process;	 

  process(CLK80)
  begin
    if (CLK80'event and CLK80='1') then
      if (sigSLatchReg='1') then
        avSigSReg(0) 
	       <= sigSReg(13) & sigSReg(13) & sigSReg(13) & sigSReg(13) & sigSReg;
        for I in 1 to 9 loop
	       avSigSReg(I) <= avSigSReg(I-1) 
		      + (sigSReg(13) & sigSReg(13) & sigSReg(13) & sigSReg(13) & sigSReg);
	     end loop;
      end if;
    end if;
  end process;	 
	
------------   END OF SIGNAL AVERAGER   ------------
	
-------------------  FIR FILTER  ------------------

process(CLK80)
begin
if (CLK80'event and CLK80='1') then
  if (DFTrigCReg='1') then
    for I in 0 to firLength-1 loop
      CQRealReg(I) <= avSigCReg(9)(17 downto 4) * DFcoeff(I);
    end loop;

      CSRealReg(0) <= CQRealReg(0)(23) & CQRealReg(0)(21 downto 0);

    for I in 1 to firLength-1 loop
      CSRealReg(I) <= CSRealReg(I-1) + (CQRealReg(I)(23) & CQRealReg(I)(21 downto 0));
    end loop;
  end if; -- 
end if; -- CLK
end process;


process(CLK80)
begin
if (CLK80'event and CLK80='1') then
  if (DFTrigSReg='1') then
    for I in 0 to firLength-1 loop
      CQImagReg(I) <= avSigSReg(9)(17 downto 4) * DFcoeff(I);
    end loop;

    CSImagReg(0) <= CQImagReg(0)(23) & CQImagReg(0)(21 downto 0);

    for I in 1 to firLength-1 loop
      CSImagReg(I) <= CSImagReg(I-1) + (CQImagReg(I)(23) & CQImagReg(I)(21 downto 0));
    end loop;
  end if; -- 
end if; -- CLK
end process;

---------------  END OF FIR FILTER -----------------


------------------------------------------------
--  signal multiplexer according to           --
--  whether digital filtration is used        --
------------------------------------------------

process(CLK80)
begin
  if (CLK80'event and CLK80='1') then
    if (DFTrigCReg='1') then sigCReg2a <= CSRealReg(134); end if;
    if (DFTrigSReg='1') then sigSReg2a <= CSImagReg(134); end if;
  end if;
end process;


process(CLK80)
begin
  if (CLK80'event and CLK80='1') then
--	   if (sigCLatchReg='1') then
	   if (C50nsLatchReg='1') then
        sigCReg2b <= sigCReg(13) & sigCReg(13) & sigCReg(13) & 
                    sigCReg(13) & sigCReg(13) & sigCReg(13) &
                    sigCReg(13) & sigCReg(13) & sigCReg(13) & sigCReg;
		end if;
--  	   if (sigSLatchReg='1') then
  	   if (S50nsLatchReg='1') then
        sigSReg2b <= sigSReg(13) & sigSReg(13) & sigSReg(13) &
                  sigSReg(13) & sigSReg(13) & sigSReg(13) &
                  sigSReg(13) & sigSReg(13) & sigSReg(13) & sigSReg;
      end if;
  end if;
end process;


with AFReg select
  sigCReg3 <= sigCReg2a when x"00000004",
              sigCReg2b when others;
  
with AFReg select
  sigSReg3 <= sigSReg2a when x"00000004",
              sigSReg2b when others;





-- ********************************************
-- ********** Signal Accumulator **************
-- ********************************************

--------------   RCVR RAM   -------------------
----   REAL component   ----
  process(CLK160,weReg)
  begin
    if (CLK160'event and CLK160 = '1') then
      if (weReg = '1') then
	     FID_real(to_integer(unsigned(addQReg))) <= QsigCReg;
      end if; -- accumTriggerReg
      AsigCReg <= FID_real(to_integer(unsigned(addQReg)));
    end if; -- CLK160
  end process;

---- IMAGINARY component ----
  process(CLK160,weReg)
  begin
    if (CLK160'event and CLK160 = '1') then
      if (weReg = '1') then
 	     FID_imag(to_integer(unsigned(addQReg))) <= QsigSReg;
      end if; -- accumTriggerReg
      AsigSReg <= FID_imag(to_integer(unsigned(addQReg)));
    end if; -- CLK80
  end process;
------------------------------


  U2: signalAccumulator 
       generic map(
         maxALBits=>maxALBits
        )         
       port map(
         CLK=>CLK160,
--         RST=>QDRSTReg_SYNC,
         RST=>samplingTriggerReg,
         RUNQ=>RUNQ,
         ACQ_PHASE=>ACQ_PHASE,
         ACQ_START=>ACQ_START,
         SIG_C=>sigCReg3,
         SIG_S=>sigSReg3,
         ASIG_C=>AsigCReg, --
         ASIG_S=>AsigSReg, --
         DWN=>dwReg(15 downto 0), 
         AL=>alReg(maxALBits-1 downto 0),
         NA=>naReg,
         STEP=>stepReg(15 downto 0),
         TRANS_ACTIVE=>transActiveReg,
         ACQ_ACTIVE=>acqActiveReg,
    --     CA=>caReg,
         QSIG_C=>QsigCReg, --
         QSIG_S=>QsigSReg, --
         ADDRESS=>add1Reg,
         WE=>weReg,
         ACCUM_OK=>accumOKReg
         );

  ACQ_BUSY <= acqActiveReg;

-- RCVR RAM Address multiplexer

with transActiveReg select addQReg <= ADD1Reg when '0',
                                      ADD2Reg when '1';

  U7: usbArbiter port map(
            CLK=>CLK20,
            RST=>RST,
            RXF_N=>RXF_N,
            TXE_N=>TXE_N,
            RD_N=>RD_N,
            WR=>WR,
            BUS_BUSY=>busBusyReg,
            TG_READY=>readyReg,
            USB_DATA=>USBDATA,
            TG_DO=>diReg,
            TG_DI=>doReg,
            TG_RD=>tgrdReg,
            TG_WR=>tgwrReg
           );

-- +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
-- +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
-- +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  RDPReg <= A(5) & A(6);
  STPReg <= A(4) & A(5) & A(6);

  TRANS_BUSY <= transActiveReg;
  

  process(CLK20) begin
  if (CLK20'event and CLK20='1') then
    if RST='1' then 
      bufLength <= 0;
      transActiveReg <= '0';
      INTF_stateReg <= x"00";
    else
      case INTF_stateReg is

    -- MAIN --
        when x"00" => 
          bufLength <= 0;
          INTF_stateReg <= x"01";
          transActiveReg <= '0';

        when x"01" =>
          readyReg <= '1';
          if (tgrdReg='1') then
            readyReg <= '0';
            INTF_stateReg <= x"02";
          else
            INTF_stateReg <= x"01";
          end if;

        when x"02" =>
          A(bufLength) <= diReg;
          bufLength <= bufLength+1;
          INTF_stateReg <= x"03";

        when x"03" =>
          if (A(bufLength-1)=X"0D") then 
            INTF_stateReg <= x"04";
          else
            INTF_stateReg <= x"05";
          end if;

        when x"04" =>
          if (bufLength=1) then
            INTF_stateReg <= x"00";
          elsif (A(0)&A(1)&A(2)&A(3)=X"53455420") then  -- "SET<SPACE>"
            INTF_stateReg <= x"1F";
          elsif (A(0)&A(1)&A(2)&A(3)&A(4)=X"5245414420") then 
            INTF_stateReg <= x"10";                     -- "READ<SPACE>"
          elsif (A(0)&A(1)=X"470D") then --"G<SPACE>"
            QDRSTReg <= '1';  -- RST signal for quad. demodulator
            INTF_stateReg <= x"F0";
          else
            INTF_stateReg <= x"0A";
          end if;

        when x"05" =>
          if (bufLength=maxComLength) then
            INTF_stateReg <= x"0A";
          else 
            INTF_stateReg <= x"01";
          end if;

    -- end of MAIN --

    -- ERROR --
        when x"0A" =>
          INTF_stateReg <= x"0B";

        when x"0B" =>
          if (busBusyReg='1') then
            INTF_stateReg <= x"0B";
          else
            doReg <= X"3F";  -- 0x3F (?)
            tgwrReg <= '1'; 
            INTF_stateReg <= x"0C";
          end if;

        when x"0C" =>
          tgwrReg <= '0';
          INTF_stateReg <= x"0D";

       when x"0D" =>
          if (busBusyReg='1') then
            INTF_stateReg <= x"0D";
          else 
            doReg <= X"0D"; -- 0x0D (<CR>)
            tgwrReg <= '1';
            INTF_stateReg <= x"0E";
          end if;

        when x"0E" => 
          tgwrReg <= '0';
          INTF_stateReg <= x"00";

    -- end of ERROR --

    -- READ --
        when x"10" =>
          case RDPReg is
            when X"414C" => -- AL
				  datBuf32Reg <= alReg;
              INTF_stateReg <= x"20";
            when X"4E41" => -- NA
				  datBuf32Reg <= naReg;
              INTF_stateReg <= x"20";
            when X"4457" => -- DW
				  datBuf32Reg <= dwReg;
              INTF_stateReg <= x"20";
            when X"5354" => -- ST
				  datBuf32Reg <= stepReg;
              INTF_stateReg <= x"20";
            when X"4146" => -- AF
				  datBuf32Reg <= afReg;
              INTF_stateReg <= x"20";
            when X"434F" => -- CO
              INTF_stateReg <= x"70";
            when others => 
              INTF_stateReg <= x"0A";
          end case;

        when x"20" =>
          kReg <= 1;
          INTF_stateReg <= x"21";
       
        when x"21" =>
          case kReg is
            when 1 => HADReg <= datBuf32Reg(31 downto 28); INTF_stateReg <= x"22";
            when 2 => HADReg <= datBuf32Reg(27 downto 24); INTF_stateReg <= x"22";
            when 3 => HADReg <= datBuf32Reg(23 downto 20); INTF_stateReg <= x"22";
            when 4 => HADReg <= datBuf32Reg(19 downto 16); INTF_stateReg <= x"22";
            when 5 => HADReg <= datBuf32Reg(15 downto 12); INTF_stateReg <= x"22";
            when 6 => HADReg <= datBuf32Reg(11 downto 8); INTF_stateReg <= x"22";
            when 7 => HADReg <= datBuf32Reg(7 downto 4); INTF_stateReg <= x"22";
            when 8 => HADReg <= datBuf32Reg(3 downto 0); INTF_stateReg <= x"22";
            when others => INTF_stateReg <= x"00";
          end case;

        when x"22" => INTF_stateReg <= x"23"; -- 1 clock delay

        when x"23" => 
          if (busBusyReg='1') then INTF_stateReg <= x"23";
          else
            doReg <= HAQReg;
            tgwrReg <= '1';
            INTF_stateReg <= x"24";
          end if;

        when x"24" => 
          tgwrReg <= '0';
          INTF_stateReg <= x"25";

        when x"25" => 
          kReg <= kReg + 1;
          INTF_stateReg <= x"21";
			

--***********************************************************************************************

        when X"70" => mReg <= 0; INTF_stateReg <= x"71";

        when X"71" => 
          if (mReg=dfLengthReg) then
            INTF_stateReg <= x"00";
          else
            INTF_stateReg <= x"72";
          end if;

        when X"72" =>
          HADReg <= std_logic_vector(dfCoeff(mReg))(9) & std_logic_vector(dfCoeff(mReg))(9) & std_logic_vector(dfCoeff(mReg))(9 downto 8);
          INTF_stateReg <= x"73";

        when x"73" => INTF_stateReg <= x"74";  -- 1 clock delay

        when X"74" =>
          if (busBusyReg='1') then INTF_stateReg <= x"74";
          else
            doReg <= HAQReg;
            tgwrReg <= '1';
            INTF_stateReg <= x"75";
          end if;

        when x"75" => 
          tgwrReg <= '0';
          INTF_stateReg <= x"76";

        when X"76" =>
          HADReg <= std_logic_vector(dfCoeff(mReg))(7 downto 4);
          INTF_stateReg <= x"77";

        when x"77" => INTF_stateReg <= x"78";  -- 1 clock delay

        when X"78" =>
          if (busBusyReg='1') then INTF_stateReg <= x"78";
          else
            doReg <= HAQReg;
            tgwrReg <= '1';
            INTF_stateReg <= x"79";
          end if;

        when x"79" => 
          tgwrReg <= '0';
          INTF_stateReg <= x"7A";

        when X"7A" =>
          HADReg <= std_logic_vector(dfCoeff(mReg))(3 downto 0);
          INTF_stateReg <= x"7B";

        when x"7B" => INTF_stateReg <= x"7C";  -- 1 clock delay

        when X"7C" =>
          if (busBusyReg='1') then INTF_stateReg <= x"7C";
          else
            doReg <= HAQReg;
            tgwrReg <= '1';
            INTF_stateReg <= x"7D";
          end if;

        when x"7D" => 
          tgwrReg <= '0';
          INTF_stateReg <= x"7E";

        when x"7E" =>
          mReg <= mReg + 1;
          INTF_stateReg <= x"71";

--***********************************************************************************************

    -- end of READ --

    -- SET --
        when x"1F" =>
          case STPReg is
            when X"414C20" => -- AL<SPACE>
              INTF_stateReg <= x"80";
            when X"4E4120" => -- NA<SPACE>
              INTF_stateReg <= x"80";
            when X"445720" => -- DW<SPACE>
              INTF_stateReg <= x"80";
            when X"535420" => -- ST<SPACE>
              INTF_stateReg <= x"80";
            when X"414620" => -- AF<SPACE>  (number of the stages in digital filter)
              INTF_stateReg <= x"80";
            when X"434F20" => -- CO<SPACE>  (coefficients for digital filter)
              INTF_stateReg <= x"D0";
            when others => 
              INTF_stateReg <= x"0A";
          end case;

        when x"80" =>
          if (bufLength=16) then
            kReg <= 1;
            INTF_stateReg <= x"81";
          else
            INTF_stateReg <= x"0A";
          end if;

        when x"81" =>
		    AHDReg <= A(6+kReg);
			 INTF_stateReg <= x"82";

        when x"82" => INTF_stateReg <= x"83";  -- 1 clock delay

        when x"83" =>
          case kReg is
            when 1 => datBuf32Reg(31 downto 28) <= AHQReg; INTF_stateReg <= x"84";
            when 2 => datBuf32Reg(27 downto 24) <= AHQReg; INTF_stateReg <= x"84";
            when 3 => datBuf32Reg(23 downto 20) <= AHQReg; INTF_stateReg <= x"84";
            when 4 => datBuf32Reg(19 downto 16) <= AHQReg; INTF_stateReg <= x"84";
            when 5 => datBuf32Reg(15 downto 12) <= AHQReg; INTF_stateReg <= x"84";
            when 6 => datBuf32Reg(11 downto 8) <= AHQReg; INTF_stateReg <= x"84";
            when 7 => datBuf32Reg(7 downto 4) <= AHQReg; INTF_stateReg <= x"84";
            when 8 => datBuf32Reg(3 downto 0) <= AHQReg; INTF_stateReg <= x"84";
				when 9 => INTF_stateReg <= x"85";
            when others => -- do nothing
          end case;

        when x"84" =>
          kReg <= kReg + 1;
          INTF_stateReg <= x"81";
			 
		  when x"85" =>
		    case STPReg is
			   when X"414C20" => -- AL<SPACE>
              alReg <= datBuf32Reg; INTF_stateReg <= x"00";
            when X"4E4120" => -- NA<SPACE>
              naReg <= datBuf32Reg; INTF_stateReg <= x"00";
            when X"445720" => -- DW<SPACE>
              dwReg <= datBuf32Reg; INTF_stateReg <= x"00";
            when X"535420" => -- ST<SPACE>
              stepReg <= datBuf32Reg; INTF_stateReg <= x"00";
            when X"414620" => -- AF<SPACE>  (number of the stages in digital filter)
              afReg <= datBuf32Reg; INTF_stateReg <= x"00";
				when others => INTF_stateReg <= x"0A";
			 end case;
          			 


        when x"D0" =>
          kReg <= 7; mReg <= 0;
          INTF_stateReg <= x"D1";

        when x"D1" =>
          AHDReg <= A(kReg);
          AHD2Reg <= A(kReg+1); 
          AHD3Reg <= A(kReg+2); 
          INTF_stateReg <= x"D2";

        when x"D2" => 
          INTF_stateReg <= x"D3";  -- 1 clock delay

        when x"D3" => 
          dfCoeff(mReg) <= signed(coReg);
          INTF_stateReg <= x"D4";

        when x"D4" =>
          if ((kReg+3)=bufLength-1 or (kReg+2)=bufLength-1 or (kReg+1)=bufLength-1) then
            dfLengthReg <= mReg+1;
            INTF_stateReg <= x"00";
          else
            INTF_stateReg <= x"D5";
          end if;

        when x"D5" => 
          kReg <= kReg + 3;
          mReg <= mReg + 1;
          INTF_stateReg <= x"D1";
 

 

    -- end of SET --

    --  RUN --
        when x"F0" => 
          QDRSTReg <= '0';
          transActiveReg <= '0';
          readyReg <= '0';
          bufLength <= 0;
          tgwrReg <= '0';
          if (RUNQ='0') then
            readyReg <= '1';
            INTF_stateReg <= x"00";
          else
            INTF_stateReg <= x"F1";
          end if;
          
        when x"F1" =>
          if (RUNQ='0') then
            INTF_stateReg <= x"00";
          elsif (accumOKReg='1') then
            transActiveReg <= '1';
            add2Reg <= (others => '0');
            INTF_stateReg <= x"F2";
          else
            INTF_stateReg <= x"F1";
          end if;

        when x"F2" =>
          kReg <= 1;
          INTF_stateReg <= x"F3";

        when x"F3" =>
          INTF_stateReg <= x"F4";

        when x"F4" =>
          case kReg is
            when 1 => doReg <= AsigCReg(31 downto 24); INTF_stateReg <= x"F5";
            when 2 => doReg <= AsigCReg(23 downto 16); INTF_stateReg <= x"F5";
            when 3 => doReg <= AsigCReg(15 downto 8); INTF_stateReg <= x"F5";
            when 4 => doReg <= AsigCReg(7 downto 0); INTF_stateReg <= x"F5";
            when 5 => doReg <= AsigSReg(31 downto 24); INTF_stateReg <= x"F5";
            when 6 => doReg <= AsigSReg(23 downto 16); INTF_stateReg <= x"F5";
            when 7 => doReg <= AsigSReg(15 downto 8); INTF_stateReg <= x"F5";
            when 8 => doReg <= AsigSReg(7 downto 0); INTF_stateReg <= x"F5";
            when others => INTF_stateReg <= x"F7";
          end case;
        
        when x"F5" =>
          if (busBusyReg='0') then 
            tgwrReg <= '1';
            kReg <= kReg + 1;
            INTF_stateReg <= x"F6";
          elsif (RUNQ='0') then
            INTF_stateReg <= x"00";
          else
            INTF_stateReg <= x"F5";
          end if;
        
        when x"F6" =>
          tgwrReg <= '0';
          INTF_stateReg <=x"F4";

        when x"F7" =>
          if (add2Reg=alReg) then
            readyReg <= '1';
            INTF_stateReg <= x"F8";
          else
            add2Reg <= add2Reg + '1';
            INTF_stateReg <= x"F2";
          end if;

       when x"F8" =>
         if (tgrdReg='1') then 
--         ackFromPCReg <= diReg;
           INTF_stateReg <= x"F0";
         elsif (RUNQ='0') then -- in case PPG is aborted
                               -- Without this, the process may fall into
           INTF_stateReg <= x"00";     -- an infinite loop here at x"F8"!!
         else
           INTF_stateReg <= x"F8";
         end if;

        
    -- end of RUN --
    
         when others => INTF_stateReg <= x"00";
      end case;
    end if; -- RST
  end if; -- CLK20
  end process;







-- +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
-- +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
-- +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++





  U_HA: hexToA port map(CLK=>CLK20,
                      RST=>RST,
                      G=>'1',
                      D=>HADReg,
                      Q=>HAQReg
                      );
  
  U_AH: aToHex port map(CLK=>CLK20,
                      RST=>RST,
                      G=>'1',
                      D=>AHDReg,
                      Q=>AHQReg);

  U_AH2: aToHex port map(CLK=>CLK20,
                      RST=>RST,
                      G=>'1',
                      D=>AHD2Reg,
                      Q=>AHQ2Reg);
          
  U_AH3: aToHex port map(CLK=>CLK20,
                      RST=>RST,
                      G=>'1',
                      D=>AHD3Reg,
                      Q=>AHQ3Reg);

  coReg <= AHQReg(1 downto 0) & AHQ2Reg & AHQ3Reg;


end RTL;
